/* ucac.c   file to read UCAC2 catalog */
/*
  Catalog is 288 strips of 0.5 degrees in dec, 0-24 hr each strip.
  UCAC numbers are sequential from first strip. 1-48,330,571
  and covers mag 8++. We need to limit data to 10 bytes per star or less
  to fit on a single CD. = 483MB
  i.e RA 4, DEC 4, MAG 2, (no ID number and no spectrum info)

  Combining strips may need care to calculate UCAC number since there
  are so many stars that it is tempting to remove the number (save 4 bytes)
  and calculate the number from a starting index number.

  Using 0.5 deg x 0.1 hr bins requires 86400 bins (whole sky if UCAC3)
  UCAC2 has 288 zones; gives 69120 bins.

  file read: ver, bin_offset[86400], UCAC_start[86400], UCAC_nstars[86400]
  strip read: [ucac_data][ucac_data][etc...]

  procedure: read file bin_offset[] etc to memory.

  to read a bin:
  calculate bin number    bin=240*(dec+90)*2+RA*10
  fseek(bin_offset[bin])
  fread(ucac_data[nstars]), calculate UCAC numbers, store in star structs

  Bright Star Supplement (BSS) has higher numbers 50,000,000++
  (430,000 stars) taken from Tycho etc with 5 degree high strips!
  IDcode 4, RA 4, DEC 4, Mag 2, SPEC 2, = 16x430000 = 5.16MB

  Using 5degrees x 0.5hr blocks needs 36x48 = 1728 blocks.
  0.1hr bins needs 5 deg * 240 = 36*240 = 8640 bins.

  BSS data file: ver, block_offset[8640], BSS_start[8640], BSS_nstars[8640],
     [bss_data][...]

  calculate BSS block     block=240*(dec+90)/5 + RA*10
  fseek(block_offset[block])
  fread(bss_data[nstars]), flag ID numbers, store in star structs

  Then plot using star structs as normal?

UCAC index file:  u2index.da
  Each 0.5deg zone is divided into 0.1hr bins. This file records the
  accumulated total stars at the end of each bin. 2 for bin = 0.
  Hence the UCAC number at the start of each bin can be calculated from
  the accumulated total of the previous bin. The number of stars in each
  bin can be found by subtracting the previous bin from the current one.

  nstars[bin] = total[bin] - total[bin-1];
  bin_offset[bin]  = total[bin-1]*length_of_data + HEADER_LENGTH;
  UCAC_start[bin] = 1 + total[bin-1];

UCAC2 Zone file format:
The following items are on each star record of 44 bytes:
(we use the first 3 items: ra dec mag)

Num Bytes  Fmt Unit       Label   Explanation
-----------------------------------------------------------------------------
 1   1- 4  I*4 mas        RA      Right Ascension at epoch J2000.0 (ICRS) (2)
 2   5- 8  I*4 mas        DE      Declination at epoch J2000.0 (ICRS)     (2)
 3   9-10  I*2 0.01 mag   U2Rmag  Internal UCAC magnitude (red bandpass)  (3)
 4  11     I*1 mas        e_RAm   s.e. at central epoch in RA (*cos DEm)(1,4)
 5  12     I*1 mas        e_DEm   s.e. at central epoch in Dec          (1,4)
 6  13     I*1            nobs    Number of UCAC observations of this star(5)
 7  14     I*1            e_pos   Error of original UCAC observ. (mas)  (1,6)
 8  15     I*1            ncat    # of catalog positions used for pmRA, pmDC
 9  16     I*1            cflg    ID of major catalogs used in pmRA, pmDE (7)
10  17-18  I*2 0.001 yr   EpRAm   Central epoch for mean RA, minus 1975   (8)
11  19-20  I*2 0.001 yr   EpDEm   Central epoch for mean DE, minus 1975   (8)
12  21-24  I*4 0.1 mas/yr pmRA    Proper motion in RA (no cos DE)         (9)
13  25-28  I*4 0.1 mas/yr pmDE    Proper motion in DE                     (9)
14  29     I*1 0.1 mas/yr e_pmRA  s.e. of pmRA (*cos DEm)                 (1)
15  30     I*1 0.1 mas/yr e_pmDE  s.e. of pmDE                            (1)
16  31     I*1 0.05       q_pmRA  Goodness of fit for pmRA             (1,11)
17  32     I*1 0.05       q_pmDE  Goodness of fit for pmDE             (1,11)
18  33-36  I*4            2m_id   2MASS pts_key star identifier          (12)
19  37-38  I*2 0.001 mag  2m_J    2MASS J  magnitude                     (13)
20  39-40  I*2 0.001 mag  2m_H    2MASS H  magnitude                     (13)
21  41-42  I*2 0.001 mag  2m_Ks   2MASS K_s magnitude                    (13)
22  43     I*1            2m_ph   2MASS modified ph_qual flag          (1,14)
23  44     I*1            2m_cc   2MASS modified cc_flg                (1,15)
-----------------------------------------------------------------------------

Example:
As an example, the data of the first 5 stars from z001 are
printed here in ASCII.  These are the numbers obtained when
using the  u2dumpz.f  program.

   item  1          2    3   4   5  6  7  8  9    10    11      12
   1246420 -322767602 1591  75  87  2 97  2  1 17330 15213   31020
   4125230 -323707012 1382  15  28  8 21  2  1 23246 22659  134316
   7345118 -322447512 1579  24  26  4 25  2  1 22903 22799   45845
   8139385 -322284308 1479  53  25  5 41  2  1 20455 22911   18125
  11128880 -323115466 1631  15  43  2 27  2  1 23364 21562   39404

 item 13  14  15  16  17         18    19    20    21  22  23
     -35  70  74  20  20 1229086517 14428 13865 13751 000 000
    -187  61  62  20  20 1101364107 12467 12131 11963 000 000
     104  61  62  20  20 1329022468 14169 13752 13708 220 000
     -16  65  61  20  20 1085341332 13111 12511 12339 000 000
     -74  61  63  20  20 1085341834 14515 13918 13877 000 000

BSS Format for binary version
(we use the first 3 items: ra dec mag)

Fld Bytes  Fmt Unit       Label   Explanation                           Notes
-----------------------------------------------------------------------------
 1   1- 4  I*4 mas        RA      Right Ascension at epoch J2000.0 (ICRS) (2)
 2   5- 8  I*4 mas        DE      Declination at epoch J2000.0 (ICRS)     (2)
 3   9-10  I*2 0.01 mag   mag     Mean magnitude                          (3)
 4  11     I*1 mas        e_RAm   S.e. at central epoch in RA (*cos DE) (1,4)
 5  12     I*1 mas        e_DEm   S.e. at central epoch in Dec          (1,4)
 6  13     I*1            magsrc  Source of magnitude                     (5)
 7  14     I*1 mas        e_pos   Error of original UCAC observ.          (6)
 8  15     I*1            ncat    # of catalog posit. for pmRA, pmDC      (6)
 9  16     I*1            cflg    ID of cats. used for astrometry fields  (7)
10  17-18  I*2 0.01 yr    EpRAm   Central epoch for mean RA, minus 1975   (8)
11  19-20  I*2 0.01 yr    EpDEm   Central epoch for mean DE, minus 1975   (8)
12  21-24  I*4 0.1 mas/yr pmRA    Proper motion in RA (no cos DE)         (9)
13  25-28  I*4 0.1 mas/yr pmDE    Proper motion in DE                     (9)
14  29     I*1 0.1 mas/yr e_pmRA  S.e. of pmRA (*cos DE)               (1,10)
15  30     I*1 0.1 mas/yr e_pmDE  S.e. of pmDE                         (1,10)
16  31     I*1 0.05       q_pmRA  Goodness of fit for pmRA             (1,11)
17  32     I*1 0.05       q_pmDE  Goodness of fit for pmDE             (1,11)
18  33-36  I*4            2m_id   2MASS pts_key star identifier          (12)
19  37-38  I*2 0.001 mag  2m_J    2MASS J magnitude                      (13)
20  39-40  I*2 0.001 mag  2m_H    2MASS H magnitude                      (13)
21  41-42  I*2 0.001 mag  2m_Ks   2MASS K_s magnitude                    (13)
22  43     I*1            2m_ph   2MASS modified ph_qual flag          (1,14)
23  44     I*1            2m_cc   2MASS modified cc_flg                (1,15)
-----------------------------------------------------------------------------

Example:
As an example, the data of the first 3 stars from s01 are
printed here in ASCII.  These are the numbers obtained when
using the  bssdumpz.f  program.

 item    1          2    3   4   5  6  7  8  9    10    11      12
   6854506 -309739533  738   1   1  0  0  0 -1  1625  1625    3290    
  11379804 -314290944 1105  40  43  0  0  0 -2   934   711   44333    
  34875774 -307484785  783   1   1  0  0  0 -1  1625  1625  -10799

 item 13  14  15  16  17         18    19    20    21  22  23
      70   6   5  20  20 1085334238  7323  7387  7360 000 000
      37  16  16  28  26 1329017546  9623  9259  9189 000 000
    -288   6   6  20  20 1181015501  6399  6068  5956 000 000


*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

//#include <allegro.h>     // delete: this is my ASTEROID compatability code.
//#include "ppmstar.h"


#define MAXZONES         288            // change to 360 for UCAC3
#define UCACBINS         MAXZONES*240   //currently 69120 UCAC2, 86400 UCAC3
#define UCAC_HEADERSIZE  (4+UCACBINS*12)
#define UCAC_DATASIZE    10

#define BSSBLOCKS         8640
#define BSS_HEADERSIZE    (4+BSSBLOCKS*12)
#define BSS_DATASIZE      16
#define BSS_STARTNUMBER   50000001      // change for UCAC3

#define UCAC_INDEX_ERR       1
#define UCAC_READ_FILE_ERR   2
#define UCAC_WRITE_FILE_ERR  3
#define BSS_INDEX_ERR        4
#define BSS_READ_FILE_ERR    5
#define BSS_WRITE_FILE_ERR   6
#define UCAC_VERSION_ERR     7
#define BSS_VERSION_ERR      8


#define MAS2HR   54000000         // 1000*60*60*15
#define MAS2DEG  3600000          // 1000*60*60

// these three items are stored in my ucac data file header after UCACBINS
int bin_offset[UCACBINS];
int UCAC_start[UCACBINS];
int UCAC_nstars[UCACBINS];


// these three items are stored in my bss data file header after BSSBLOCKS
int block_offset[BSSBLOCKS];
int BSS_start[BSSBLOCKS];          // 50,000,001++
int BSS_nstars[BSSBLOCKS];


char ucacfile[80] = {"u2\\"};       // default UCAC sub directory
char bssfile[80] = {"bss\\"};       // I don't use this...

FILE *fpout;              // global output file for data conversion


// read UCAC index data:    2 4 8 10 11 etc

int read_UCAC_index(char *fname)
{
   int cx, bin, oldbin = 0;
   FILE *fp;

   fp = fopen(fname, "rb");
   if (!fp)
   {
      return UCAC_INDEX_ERR;
   }
   for (cx = 0; cx < UCACBINS; cx++)
   {
      fread(&bin, sizeof(int), 1,fp);
      bin_offset[cx] = UCAC_HEADERSIZE + oldbin*UCAC_DATASIZE;
      UCAC_start[cx] = 1 + oldbin;
      UCAC_nstars[cx] = bin - oldbin;
      oldbin = bin;      // remember last value of running total.
   }
   fclose(fp);
   return 0;
}


// read BSS index data:    0 1 2 2 etc

int read_BSS_index(char *fname)
{
   int cx, bin, oldbin = 0;
   FILE *fp;

   fp = fopen(fname, "rb");
   if (!fp)
   {
      return BSS_INDEX_ERR;
   }
   for (cx = 0; cx < BSSBLOCKS; cx++)
   {
      fread(&bin, sizeof(int), 1,fp);
      block_offset[cx] = BSS_HEADERSIZE + oldbin*BSS_DATASIZE;
      BSS_start[cx] = BSS_STARTNUMBER + oldbin;
      BSS_nstars[cx] = bin - oldbin;
      oldbin = bin;      // remember last value of running total.
   }
   fclose(fp);
   return 0;
}


// convert UCAC to star file, stores 10 bytes RA 4, DEC 4, MAG 2
// converts milli-arcsec to floating point for RA and Dec.
// converting RA and DEC to floating point saves converting when read later.
// we also use this to store some BSS data.
// Change this to suit your own code...

int store_UCAC_file(int masra, int masdec, int mag)
{
   float ra, dec;
   short m;

   ra = (float)masra / MAS2HR;
   dec = (float)masdec / MAS2DEG;
   m   = mag;                              // removed / 10; 
   fwrite(&ra,  sizeof(float), 1, fpout);
   fwrite(&dec, sizeof(float), 1, fpout);
   fwrite(&m,   sizeof(short), 1, fpout);
   return 0;
}


// read ucac zone file data:    stars are in 44 byte blocks
// fname is the filename including PATH
// num   is the number of stars to read (obtained from index file)

int read_UCAC_zone(char *fname, int num)
{
   int cx, masra, masdec, mag;
   int mas[11];
   FILE *fp;

   fp = fopen(fname, "rb");
   if (!fp)
   {
      return UCAC_READ_FILE_ERR;
   }
   for (cx = 0; cx < num; cx++)
   {
      fread(mas, sizeof(mas), 1,fp);
      masra  = mas[0];
      masdec = mas[1];
      mag = mas[2] & 0xFFFF;                      // only lower 16 bits
      // store in data file
      store_UCAC_file(masra, masdec, mag);
   }
   fclose(fp);
   return 0;
}


// This function reads UCAC zones from the hard disk or from 3 CDs
// and generates the 484MB data file.
// cdid    CD ID number: 0, all files on HD; 1 2 3 CDROM disk

int write_UCAC_file(int cdid)
{
   int cx, cy, x;
   int min = 1, max = 288;           // change to 360 for UCAC3
   long tot = 0L;
   char fname[100];

   if (cdid == 1)  max = 106;        // adjust for each CD
   if (cdid == 2)
   {
      min = 107;
      max = 182;   // inclusive
   }
   if (cdid == 3)  min = 183;

   // ucacfile contains the path to u2 data i.e "D:\u2\" NOTE end seperator!

   sprintf(fname, "%s%s", ucacfile, "u2index.da");

   x = read_UCAC_index(fname);
   if (x) return x;                  // bail out if no index

   if (cdid > 1)
      fpout = fopen("ucac.dat", "ab");
   else
      fpout = fopen("ucac.dat", "wb");

   if (!fpout) return UCAC_WRITE_FILE_ERR;    // can't find output file...

   if (cdid <= 1)    // disk one: initialze file header; 2,3 just add stars.
   {
      x = UCACBINS;
      fwrite(&x, sizeof(int), 1, fpout);
      fwrite(bin_offset,  sizeof(bin_offset),  1, fpout);
      fwrite(UCAC_start,  sizeof(UCAC_start),  1, fpout);
      fwrite(UCAC_nstars, sizeof(UCAC_nstars), 1, fpout);
   }
   for (cx = min; cx < max+1; cx++)
   {
      x = 0;
      sprintf(fname, "%sz%03d", ucacfile, cx);
      for (cy = 0; cy < 240; cy++)
      {
         x += UCAC_nstars[(cx-1)*240 + cy];  // number of stars in zone
      }
      if (read_UCAC_zone(fname, x))     // read zone, write stars found
      {
         fclose(fpout);               // bail out if file missing
         printf("missing file %s\n", fname);
         printf("Found %ld UCAC stars \n", tot);    // sum up
         return UCAC_READ_FILE_ERR;
      }
      printf("%6d in %s   ", x, fname);   // report progress...
      tot += x;
   }
   fclose(fpout);
   printf("\nFound %ld UCAC stars \n", tot);
   return 0;
}


// Two functions to show how to use the file we just made...

// read UCAC file header, offset etc

int read_UCAC_fileheader()
{
   int x;

   fpout = fopen("ucac.dat", "rb");
   if (!fpout) return UCAC_READ_FILE_ERR;

   // read UCACBINS from file as a version check

   fread(&x, sizeof(int), 1, fpout);
   if (x != UCACBINS)
   {
      fclose(fpout);
      return UCAC_VERSION_ERR;
   }

   fread(bin_offset,  sizeof(bin_offset),  1, fpout);
   fread(UCAC_start,  sizeof(UCAC_start),  1, fpout);
   fread(UCAC_nstars, sizeof(UCAC_nstars), 1, fpout);

   fclose(fpout);
   return 0;
}


// read UCAC file bin for stars
// bin:   number of bin to read 0 to UCACBINS-1
// idx:   index to an array where you put the data
// (my implementation is commented out: replace with your own code)

int read_UCAC_filebin(int bin, int idx)
{
   int cx, start;
   float ra, dec;
   short m;

   fpout = fopen("ucac.dat", "rb");
   if (!fpout) return UCAC_READ_FILE_ERR;

   fseek(fpout, bin_offset[bin], SEEK_SET);

   start = UCAC_start[bin];

   for (cx = 0; cx < UCAC_nstars[bin]; cx++)
   {
      fread(&ra, sizeof(float), 1, fpout);
      fread(&dec, sizeof(float), 1, fpout);
      fread(&m, sizeof(short), 1, fpout);
//      star[idx].num = start++;
//      star[idx].ra = ra;
//      star[idx].dec = dec;
//      star[idx].mag = m;
//      star[idx].sp[0] = 'G';
//      star[idx].sp[1] = '5';
//      calcstardata(idx);
      idx++;
   }

   fclose(fpout);
   return 0;
}

// BRIGHT STAR SUPPLEMENT

// read bss zone file data:    stars are in 44 byte blocks
// fname:   filename including path i.e.  "bss\s01"
// num:     number of stars to read (from BSS_nstars[])
// zone:    zone being searched, 0="bss\s01" ... 35="bsss\s36"
// NOTE: uses bogus spectrum info; replace later with tycho2 colour index...

int read_BSS_zone(char *fname, int num, int zone)
{
   int cx, masra, masdec, mag;
   int mas[11];
   int number;
   char spec[] = "G5";    // bogus spectrum for compatability with ASTEROID.
   FILE *fp;

   fp = fopen(fname, "rb");
   if (!fp)
   {
      return BSS_READ_FILE_ERR;
   }

   // UCAC2 star id number
   number = BSS_start[zone * 240];

   for (cx = 0; cx < num; cx++)
   {
      fread(mas, sizeof(mas), 1,fp);
      masra  = mas[0];
      masdec = mas[1];
      mag    = mas[2] & 0xFFFF;                   // only lower 16 bits
      // store in data file
      fwrite(&number, sizeof(int), 1, fpout);     // 4 bytes
      store_UCAC_file(masra, masdec, mag);        // convert 10 bytes
      fwrite(spec, 2, 1, fpout);                  // 2 bytes = 16 each star
      number++;
   }
   fclose(fp);
   return 0;
}


// This function generates the BSS data file.
// All BSS data is expected in a sub-directory called "bss"

int write_BSS_file()
{
   int cx, cy, x;
   long tot = 0L;
   char fname[100];

   x = read_BSS_index("bss\\bsindex.da");
   if (x) return x;

   fpout = fopen("bssucac.dat", "wb");
   if (!fpout) return BSS_WRITE_FILE_ERR;

   x = BSSBLOCKS;
   fwrite(&x, sizeof(int), 1, fpout);
   fwrite(block_offset, sizeof(block_offset), 1, fpout);
   fwrite(BSS_start, sizeof(BSS_start), 1, fpout);
   fwrite(BSS_nstars, sizeof(BSS_nstars), 1, fpout);
   x = 0;
   for (cx = 0; cx < 36; cx++)
   {
      x = 0;
      sprintf(fname, "bss\\s%02d", cx+1);
      for (cy = 0; cy < 240; cy++)
      {
         x += BSS_nstars[cx*240 + cy];
      }
      printf("reading %d stars from %s\n", x, fname);
      read_BSS_zone(fname, x, cx);     // read zone, write stars found
      tot += x;
   }
   fclose(fpout);
   printf("read %ld BSS stars \n", tot);
   return 0;
}


// Two example functions to access BSS data file.

// read BSS file header, offset etc

int read_BSS_fileheader()
{
   int x;

   fpout = fopen("bssucac.dat", "rb");
   if (!fpout) return BSS_READ_FILE_ERR;

   // x = UCACBINS;
   fread(&x, sizeof(int), 1, fpout);
   if (x != BSSBLOCKS)
   {
      fclose(fpout);
      return BSS_VERSION_ERR;
   }

   fread(block_offset, sizeof(block_offset), 1, fpout);
   fread(BSS_start, sizeof(BSS_start), 1, fpout);
   fread(BSS_nstars, sizeof(bin_offset), 1, fpout);

   fclose(fpout);
   return 0;
}



// read BSS file bin for stars
// bin:   number of bin to read 0 to BSSBLOCKS-1
// idx:   index to an array where you put the data
// (my implementation is commented out: replace with your own code)

int read_BSS_filebin(int bin, int idx)
{
   int cx, number;
   float ra, dec;
   short m;
   char spec[2];

   fpout = fopen("bssucac.dat", "rb");
   if (!fpout) return BSS_READ_FILE_ERR;

   fseek(fpout, block_offset[bin], SEEK_SET);

   // read 16 bytes per star;

   for (cx = 0; cx < BSS_nstars[bin]; cx++)
   {
      fread(&number, sizeof(int),   1, fpout);
      fread(&ra,     sizeof(float), 1, fpout);
      fread(&dec,    sizeof(float), 1, fpout);
      fread(&m,      sizeof(short), 1, fpout);
      fread(spec,    sizeof(spec),  1, fpout);
//      star[idx].num = number;
//      star[idx].ra = ra;
//      star[idx].dec = dec;     // replace with your code...
//      star[idx].mag = m;
//      star[idx].sp[0] = spec[0];
//      star[idx].sp[1] = spec[1];
//      calcstardata(idx);
      idx++;
   }

   fclose(fpout);
   return 0;
}



// test code to generate the data files...

// If all files copied to hard disk:
//   install in parent directory and run ucac with no parameters.
//      (ucac expects index and all zone files in the "u2" sub directory)
// If using the cdroms:
//   install in any directory and run ucac with explicit parameters
//   ucac [dir] [cd_number] [bss flag]
//         dir  cdrom directory with zone files, eg: "D:\u2\"
//              (note that the last character MUST be a file seperator)
//         cd number:   use 1, 2 or 3 for the ucac cd in the drive
//            (always run with cd 1 first, follow with 2 then 3 last.)
//         bss flag: only use if BSS data is in a BSS sub directory.

int main(int argc, char *argv[])
{
   int cx, cs = 0;
   int cdid = 0;   // all ucac2 files on hard disk
   int bss = 0;    // don't look for bss files

   if (argc > 1)  strcpy(ucacfile, argv[1]);    // data directory
   if (argc > 2)  cdid = atoi(argv[2]);         // cd id code
   if (argc > 3)  bss  = atoi(argv[3]);         // bss flag

   printf("UCAC [directory] [cdnumber] [bss flag]\n");
   printf("UCAC %s %d %d\n", ucacfile, cdid, bss);

   if (bss)
   {
      cs = write_BSS_file();    // look for bss data

      switch (cs)
      {
      case BSS_INDEX_ERR:        
         printf("BSS Index file error\n"); break;
      case BSS_READ_FILE_ERR:    
         printf("BSS Read file error\n"); break;
      case BSS_WRITE_FILE_ERR:   
         printf("BSS Write file error\n"); break;
      case BSS_VERSION_ERR:      
         printf("BSS format error\n"); break;
      }
   }

   cx = write_UCAC_file(cdid);

   switch (cx)
   {
   case UCAC_INDEX_ERR:
      printf("UCAC Index file error\n"); break;
   case UCAC_READ_FILE_ERR:   
      printf("UCAC File read error\n"); break;
   case UCAC_WRITE_FILE_ERR:  
      printf("UCAC File write error\n"); break;
   case UCAC_VERSION_ERR:     
      printf("UCAC Version error\n"); break;
   }
   return 0;
}




/* end code */



