/* ucac.c   file to read UCAC3 catalog */
/*
  Catalog is 360 strips of 0.5 degrees in dec, 0-24 hr each strip.
  UCAC numbers are sequential from first strip. 1-100,766,462
  and covers mag 8++.

  Using 0.5 deg x 0.1 hr bins requires 86400 bins (whole sky if UCAC3)

  file read: bins, bin_offset[86400], UCAC_start[86400], UCAC_nstars[86400]
  strip read: [ucac_data][ucac_data][etc...]

  procedure: read file bin_offset[] etc to memory.

  to read a bin:
  calculate bin number    bin=240*(dec+90)*2+RA*10
  fseek(bin_offset[bin])
  fread(ucac_data[nstars]), calculate UCAC numbers, store in star structs

UCAC3 index file:  u3index.asc
  Each 0.5deg zone is divided into 0.1hr bins. This file records the
  the UCAC number in each bin. A running total can be calculated from this.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

//#include <allegro.h>     // delete: this is my ASTEROID compatability code.
//#include "ppmstar.h"
//#include "astbin.h"

#define MAXZONES         288              // change to 360 for UCAC3
#define UCACBINS         (MAXZONES*240)   //currently 69120 UCAC2, 86400 UCAC3
#define UCAC3BINS        86400
#define UCAC_HEADERSIZE  (4+UCACBINS*12)
#define UCAC3_HEADERSIZE  (4+UCAC3BINS*12)
#define UCAC_DATASIZE    10

#define UCAC_INDEX_ERR       1
#define UCAC_READ_FILE_ERR   2
#define UCAC_WRITE_FILE_ERR  3
#define UCAC_VERSION_ERR     7


#define MAS2HR   54000000         // 1000*60*60*15
#define MAS2DEG  3600000          // 1000*60*60

// these three items are stored in my ucac data file header after UCAC3BINS
int bin_offset[UCAC3BINS];
int UCAC_start[UCAC3BINS];
int UCAC_nstars[UCAC3BINS];

// number of bins, to determine ucac2 or ucac3

int ucacbins = 0;

int UCACflag;                       // if UCAC files found, = largest bin size

char ucacfile[80] = {"u2/"};       // default UCAC2 sub directory on CD
char ucacdatfile[256];             // set up by init_ucac()

//static char buf[256];

FILE *fpout;              // global output file for data conversion


char ufreadbuf[1000];

int ufgetaline(FILE *fp)   /* get an input line into buf and skip over newlines */
{
    char c = 0;
    char *buff;

    buff = ufreadbuf;
    *buff = 0;
    while((c=fgetc(fp)) != 0x0a)
    {
	if (ferror(fp))    return 0;
	if (feof(fp))
	{
	   *buff = 0;
	   return 0;
	}
	if (c == 0x0D) c = 0;         /* was fgetc(fp); */
	else
	 *buff++ = c;
    }
    *buff = 0;
    /* c = fgetc(fp);          skip over '\n' */
    if (feof(fp))    return 0;
    if (*ufreadbuf == 0) return 1;      // blank line 
    else           return *ufreadbuf;
}


// read UCAC3 index data:    2 4 8 10 11 etc

int read_UCAC3_index(char *fname)
{
   int cx, x, bin, oldbin = 0;
   FILE *fp;

   fp = fopen(fname, "rb");
   if (!fp)
   {
      return UCAC_INDEX_ERR;
   }
   for (cx = 0; cx < UCAC3BINS; cx++)
   {
      ufgetaline(fp);
       // get second value on line (number in each bin)
      sscanf(ufreadbuf, "%d %d", &x, &bin);
      bin_offset[cx] = UCAC3_HEADERSIZE + oldbin*UCAC_DATASIZE;
      UCAC_start[cx] = 1 + oldbin;
      UCAC_nstars[cx] = bin; // - oldbin;
      oldbin += bin;      // remember last value of running total.
   }
   fclose(fp);
   return 0;
}


// convert UCAC to star file, stores 10 bytes RA 4, DEC 4, MAG 2
// converts milli-arcsec to floating point for RA and Dec.
// converting RA and DEC to floating point saves converting when read later.
// we also use this to store some BSS data.
// Change this to suit your own code...

int store_UCAC_file(int masra, int masdec, int mag)
{
   float ra, dec;
   short m;

   ra = (float)masra / MAS2HR;
   dec = (float)masdec / MAS2DEG;
   m   = mag;                              // removed / 10; 
   fwrite(&ra,  sizeof(float), 1, fpout);
   fwrite(&dec, sizeof(float), 1, fpout);
   fwrite(&m,   sizeof(short), 1, fpout);
   return 0;
}


// read ucac3 zone file data:    stars are in 84 byte blocks
// fname is the filename including PATH
// num   is the number of stars to read (obtained from index file)

int read_UCAC3_zone(char *fname, int num)
{
   int cx, masra, masdec, mag;
   int mas[21];
   FILE *fp;

   fp = fopen(fname, "rb");
   if (!fp)
   {
      return UCAC_READ_FILE_ERR;
   }
   for (cx = 0; cx < num; cx++)
   {
      fread(mas, sizeof(mas), 1,fp);
      masra  = mas[0];
      masdec = mas[1] - 324000000;        // convert spd to dec (mas)
      mag = mas[2] & 0xFFFF;                      // only lower 16 bits
      // store in data file
      store_UCAC_file(masra, masdec, mag);
   }
   fclose(fp);
   return 0;
}




// This function reads UCAC zones from the hard disk or from DVDs
// and generates the 1008MB data file.
// cdid    CD ID number: 0, all files on HD; 1 2 DVD disk

int write_UCAC3_file(int cdid)
{
   int cx, cy, x;
   int min = 1, max = 360;           // change to 360 for UCAC3
   long tot = 0L;
   char fname[100];

   if (cdid == 1)  max = 144;        // adjust for each DVD side
   if (cdid == 2)
   {
      min = 145;
      max = 360;   // inclusive
   }

   // ucacfile contains the path to ucac3 data i.e "D:\" NOTE end seperator!

   x = read_UCAC3_index("u3index.asc");  //fname);
   if (x) return x;                  // bail out if no index

   if (cdid > 1)
      fpout = fopen("ucac.dat", "ab");
   else
      fpout = fopen("ucac.dat", "wb");

   if (!fpout) return UCAC_WRITE_FILE_ERR;    // can't find output file...

   if (cdid <= 1)    // disk one: initialze file header; 2,3 just add stars.
   {
      x = UCAC3BINS;
      fwrite(&x, sizeof(int), 1, fpout);
      fwrite(bin_offset,  sizeof(bin_offset),  1, fpout);
      fwrite(UCAC_start,  sizeof(UCAC_start),  1, fpout);
      fwrite(UCAC_nstars, sizeof(UCAC_nstars), 1, fpout);
   }
   for (cx = min; cx < max+1; cx++)
   {
      x = 0;
      sprintf(fname, "z%03d", cx);
      // find number of stars in this zone
      for (cy = 0; cy < 240; cy++)
      {
         x += UCAC_nstars[(cx-1)*240 + cy];  // number of stars in zone
      }
      if (read_UCAC3_zone(fname, x))     // read zone, write stars found
      {
         fclose(fpout);               // bail out if file missing
         printf("missing file %s\n", fname);
         printf("Found %ld UCAC stars \n", tot);    // sum up
         return UCAC_READ_FILE_ERR;
      }
      printf("%6d in %s   ", x, fname);   // report progress...
      tot += x;
   }
   fclose(fpout);
   printf("\nFound %ld UCAC stars \n", tot);
   return 0;
}

// Two functions to show how to use the file we just made...

// read UCAC file header, offset etc

int read_UCAC_fileheader()
{
   int x;

   // locate data file
   fpout = fopen(ucacdatfile, "rb");
   if (!fpout) return UCAC_READ_FILE_ERR;

   // read UCACBINS from file as a version check

   fread(&x, sizeof(int), 1, fpout);
   if (x != UCACBINS && x != UCAC3BINS)
   {
      fclose(fpout);
      return UCAC_VERSION_ERR;
   }
   ucacbins = x;
   fread(bin_offset,  sizeof(int) * ucacbins, 1, fpout);
   fread(UCAC_start,  sizeof(int) * ucacbins, 1, fpout);
   fread(UCAC_nstars, sizeof(int) * ucacbins, 1, fpout);

   fclose(fpout);
   return 0;
}


// read UCAC file bin for stars
// bin:   number of bin to read 0 to UCACBINS-1
// idx:   index to an array where you put the data
// (my implementation is commented out: replace with your own code)

int read_UCAC_filebin(int bin, int idx)
{
   int cx, start;
   float ra, dec;
   short m;

   // locate ucac2 data
   fpout = fopen(ucacdatfile, "rb");
   if (!fpout) return UCAC_READ_FILE_ERR;

   fseek(fpout, bin_offset[bin], SEEK_SET);

   start = UCAC_start[bin];

   for (cx = 0; cx < UCAC_nstars[bin]; cx++)
   {
      fread(&ra, sizeof(float), 1, fpout);
      fread(&dec, sizeof(float), 1, fpout);
      fread(&m, sizeof(short), 1, fpout);
//      star[idx].ppmnum = start++;
//      star[idx].ra = ra;
//      star[idx].dec = dec;
//      star[idx].mag = m;
//      star[idx].sp[0] = 'G';
//      star[idx].sp[1] = '5';
//      calcstardata(idx);
      idx++;
   }

   fclose(fpout);
   return 0;
}



// initialize UCAC headers and offsets etc
// returns number of stars in the largest bin

int init_ucac()
{
   int cx, x;

   UCACflag = 0;
//   getfilename(ucacdatfile, "ucac/ucac.dat");

   x = read_UCAC_fileheader();
   if (x) return x;

//   x = read_BSS_fileheader();
//   if (x) return x;

   UCACflag = 1;
   // find largest bin size
   for (cx = 0; cx < ucacbins; cx++)
   {
      if (x < UCAC_nstars[cx])
      {
         x = UCAC_nstars[cx];
         UCACflag = cx;
      }
   }

   return x;
}



// test code to generate the data files...

// If all files copied to hard disk:
//   install in parent directory and run ucac with no parameters.
//      (ucac expects index and all zone files in the "u2" sub directory)
// If using the cdroms:
//   install in any directory and run ucac with explicit parameters
//   ucac [dir] [cd_number] [bss flag]
//         dir  cdrom directory with zone files, eg: "D:\u2\"
//              (note that the last character MUST be a file seperator)
//         cd number:   use 1, 2 or 3 for the ucac cd in the drive
//            (always run with cd 1 first, follow with 2 then 3 last.)
//         bss flag: only use if BSS data is in a BSS sub directory.


int main(int argc, char *argv[])
{
   int cx, cs = 0;
   int cdid = 0;   // all ucac2 files on hard disk

   if (argc > 1)  strcpy(ucacfile, argv[1]);    // data directory
   if (argc > 2)  cdid = atoi(argv[2]);         // cd id code

   printf("UCAC [directory] [cdnumber]\n");
   printf("UCAC %s %d \n", ucacfile, cdid);

   cx = write_UCAC3_file(cdid);

   switch (cx)
   {
   case UCAC_INDEX_ERR:
      printf("UCAC Index file error\n"); break;
   case UCAC_READ_FILE_ERR:   
      printf("UCAC File read error\n"); break;
   case UCAC_WRITE_FILE_ERR:  
      printf("UCAC File write error\n"); break;
   case UCAC_VERSION_ERR:     
      printf("UCAC Version error\n"); break;
   }
   return 0;
}




/* end code */



